import inspect
import os
import re
import time
from pathlib import Path
from types import FunctionType, ModuleType

import click
from gutt.parser import load_module_from_pyfile
from gutt.template import populate_testclass, populate_testfunc
from gutt.utils import (
    blacking,
    collect_classes_and_functions,
    expand_sys_path,
    isorting,
    load_module_by_name,
    makefile,
    qualname,
)

from ..code import CodeBlock, indent


class InvalidModule(Exception):
    pass


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    from gutt import __version__

    click.echo(f"Version: {__version__}")

    ctx.exit()


def collect_items_from_module(module: ModuleType, exclude=None, sleep_interval=0.01):
    mod_impl_mappings = {}
    items = 0
    should_nl = True

    for obj in collect_classes_and_functions(module):

        obj_name = qualname(obj)
        if isinstance(exclude, str) and re.search(exclude, obj_name):
            if should_nl:
                click.echo()
                should_nl = False
            click.secho("excluding ", nl=False, fg="bright_white")
            click.secho(obj_name, fg="bright_black")

            continue

        _modname = obj.__module__

        if _modname not in mod_impl_mappings:
            mod_impl_mappings[_modname] = []

        if obj not in mod_impl_mappings[_modname]:
            items += 1
            mod_impl_mappings[_modname].append(obj)
            click.echo("\033[K", nl=False)
            click.secho(f"collecting {items} items: ", nl=False, fg="bright_white")
            click.secho(f"{obj_name}\r", nl=False, fg="bright_cyan")
            time.sleep(sleep_interval)
            should_nl = True

    click.echo()

    return mod_impl_mappings


@click.command()
@click.pass_context
@click.option(
    "--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True
)
@click.option(
    "--modname", "-m", help="Target module name for generating test templates"
)
@click.option(
    "--path",
    "-p",
    multiple=True,
    default=[os.getcwd()],
    help='Insert into "sys.path" to search modules, could assign with multiple values',
)
@click.option(
    "--exclude",
    "-e",
    help="Giving regex pattern to match the implementation to be excluded by its qualname",
)
@click.option(
    "--output",
    "-o",
    default="tests/_gutt",
    help="Output root directory for populating test files, default: tests/_gutt",
)
def main(ctx, modname, path, exclude, output):

    with expand_sys_path(*path):
        try:
            module = load_module_by_name(modname)
        except Exception:
            raise InvalidModule(f'got: "{modname}"')

        ispkg = hasattr(module, "__path__")
        mod_impl_mappings = collect_items_from_module(module, exclude=exclude)

    for mod, imps in mod_impl_mappings.items():

        if ispkg:
            mod = re.sub(rf"^{modname}\.", "", mod)
            pfx, name = mod.rsplit(".", 1) if "." in mod else ("", mod)
            pfx = os.path.join(modname.replace(".", "_"), pfx.replace(".", "/"))
        else:
            name = mod.replace(".", "_")
            pfx = ""

        filename = f"test_{name}.py"

        fullpath = os.path.join(output, pfx, filename)

        makefile(fullpath)

        _module = load_module_from_pyfile(f"test_{name}", fullpath)

        try:
            source = inspect.getsource(_module)
            click.secho("loading existing codes from ", nl=False, fg="bright_white")
            click.secho(f"{fullpath}", fg="bright_green")
        except OSError:
            source = ""

        blocks = CodeBlock.collect_from_source(source)

        implemented = {
            bk.name: i for i, bk in enumerate(blocks) if bk.kind in ("def", "class")
        }

        code_added = 0

        for obj in imps:
            fullname = qualname(obj)
            name = obj.__name__

            if isinstance(obj, FunctionType) and f"test_{name}" not in implemented:

                blocks.append(
                    CodeBlock(
                        raw=populate_testfunc(obj), name=f"test_{name}", kind="def"
                    )
                )

                code_added += 1

                click.echo("\033[K", nl=False)
                click.secho("adding test function: ", nl=False, fg="bright_white")
                click.secho(f"{fullname}\r", nl=False, fg="bright_cyan")

            elif isinstance(obj, type):

                ut_name = f"Test{name}"

                if ut_name not in implemented:
                    blocks.append(
                        CodeBlock(
                            raw=populate_testclass(obj), name=ut_name, kind="class"
                        )
                    )

                    code_added += 1

                    click.echo("\033[K", nl=False)
                    click.secho("adding test class: ", nl=False, fg="bright_white")
                    click.secho(f"{fullname}\r", nl=False, fg="bright_cyan")

                else:
                    block = blocks[implemented[ut_name]]

                    methods_implemented = {
                        bk.name: i
                        for i, bk in enumerate(block.children)
                        if bk.kind in ("def",)
                    }

                    methods_to_add = []

                    for k, v in obj.__dict__.items():
                        method_name = f"test_{k}"
                        if (
                            (
                                inspect.ismethod(v)
                                or isinstance(
                                    v, (FunctionType, classmethod, staticmethod)
                                )
                            )
                            and (not k.startswith("__"))
                            and (method_name not in methods_implemented)
                        ):

                            code = "\n".join(
                                [
                                    indent(f"def {method_name}(self):", level=1),
                                    indent("pass", level=2),
                                    "",
                                ]
                            )

                            methods_to_add.append(code)

                            click.echo("\033[K", nl=False)
                            click.secho(
                                "adding test method: ", nl=False, fg="bright_white"
                            )
                            click.secho(f"{fullname}\r", nl=False, fg="bright_cyan")

                    if methods_to_add:

                        code = "\n".join([block.raw] + methods_to_add)

                        blocks[implemented[ut_name]] = CodeBlock(
                            raw=code, kind="class", name=ut_name
                        )

                        code_added += 1

            time.sleep(0.01)

        if code_added > 0:
            click.echo()
            new_source = "\n".join([b.raw for b in blocks])
            formatted = isorting(blacking(new_source))

            makefile(fullpath, formatted, overwrite=True)
            click.secho("writing codes to ", nl=False, fg="bright_white")
            click.secho(f"{fullpath}", fg="bright_green")

        else:
            click.secho("all templates populated, skip", fg="bright_black")

    output_root = Path(output)
    paths = {output_root}
    for path in output_root.glob("**/*.py"):
        if path.name == "__init__.py":
            continue

        parent = path.parent

        while parent > output_root:

            if not parent.joinpath("__init__.py").exists():
                paths.add(parent)

            parent = parent.parent

    for p in paths:
        makefile(os.path.join(str(p), "__init__.py"))



import inspect
from types import FunctionType

from .code import indent


def populate_testclass(cls: type) -> str:
    """"""

    module = cls.__module__
    name = cls.__name__
    skeleton = [
        f"class Test{name}:",
        "",
        indent("@classmethod", level=1),
        indent("def setup_class(cls):", level=1),
        indent(f"from {module} import {name}", level=2),
        indent(f"assert {name}", level=2),
        "",
        indent("@classmethod", level=1),
        indent("def teardown_class(cls):", level=1),
        indent("pass", level=2),
        "",
        indent("def setup_method(self, method):", level=1),
        indent("pass", level=2),
        "",
        indent("def teardown_method(self, method):", level=1),
        indent("pass", level=2),
        "",
    ]

    for k, v in cls.__dict__.items():
        if (
            inspect.ismethod(v)
            or isinstance(v, (FunctionType, classmethod, staticmethod))
        ) and not k.startswith("__"):
            skeleton.extend(
                [indent(f"def test_{k}(self):", level=1), indent("pass", level=2), ""]
            )

    return "\n".join(skeleton)


def populate_testfunc(func: FunctionType) -> str:

    module = func.__module__
    name = func.__name__
    skeleton = [
        f"def test_{name}():",
        indent(f"from {module} import {name}", level=1),
        "",
        indent(f"assert {name}", level=1),
    ]

    return "\n".join(skeleton)



import io
import os
import pathlib
import pkgutil
import subprocess as sp
import sys
import tempfile
from contextlib import contextmanager
from importlib._bootstrap_external import SourceFileLoader
from itertools import tee
from types import FunctionType, ModuleType
from typing import Generator, Union


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""

    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def qualname(obj: Union[FunctionType, ModuleType, type], level: int = -1) -> str:
    """Return the qualname of a class, a function or a module.

    >>> class Foo:
    ...     pass

    >>> def bar():
    ...     pass

    >>> qualname(Foo)
    'gutt.utils.Foo'

    >>> qualname(bar)
    'gutt.utils.bar'
    """

    if isinstance(obj, FunctionType) or isinstance(obj, type):
        name = f"{obj.__module__}.{obj.__qualname__}"

    elif isinstance(obj, ModuleType):
        name = f"{obj.__name__}"

    else:
        name = ""

    return ".".join(name.split(".")[-level:]) if level > 0 else name


def load_module_by_name(modname: str) -> ModuleType:
    loader: SourceFileLoader = pkgutil.get_loader(modname)
    return loader.load_module(modname)


def collect_classes_and_functions(
    module: ModuleType,
) -> Generator[Union[FunctionType, type], None, None]:

    modname = qualname(module)
    cached = set()

    impls = (
        _collect_from_package(module)
        if hasattr(
            module, "__path__"
        )  # By definition, if a module has a __path__ attribute, it is a package.
        else _collect_from_module(module)
    )

    for obj in impls:
        name = qualname(obj)
        if obj.__module__.startswith(modname) and (name not in cached):
            cached.add(name)
            yield obj


def _collect_from_package(
    package: ModuleType,
) -> Generator[Union[FunctionType, type], None, None]:

    for obj in _collect_from_module(package):
        yield obj

    for finder, name, _ in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):

        with escape_any_commandline_parser():
            try:
                mod = finder.find_module(name).load_module(name)

            except BaseException:
                continue

        for obj in _collect_from_module(mod):

            yield obj


def _collect_from_module(
    module: ModuleType,
) -> Generator[Union[FunctionType, type], None, None]:
    for _, obj in module.__dict__.items():
        if isinstance(obj, (type, FunctionType)):

            yield obj


def blacking(source_code: str):

    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    p = sp.Popen(f"cat {fname}".split(), stdout=sp.PIPE)

    out = sp.check_output("black -q -".split(), stdin=p.stdout)

    p.wait()

    try:
        pathlib.Path(fname).unlink()
    except FileNotFoundError:
        pass

    return out.decode()


def isorting(source_code: str):

    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(source_code)
        fname = f.name

    p = sp.Popen(f"cat {fname}".split(), stdout=sp.PIPE)

    out = sp.check_output("isort -q -".split(), stdin=p.stdout)

    p.wait()

    try:
        pathlib.Path(fname).unlink()
    except FileNotFoundError:
        pass

    return out.decode()


@contextmanager
def expand_sys_path(*paths: str):

    num = len(paths)

    for p in paths[::-1]:
        if p and isinstance(p, str):

            sys.path.insert(0, p)

    yield

    for _ in range(num):
        sys.path.pop(0)


@contextmanager
def escape_any_commandline_parser():

    argv = sys.argv
    stdout = sys.stdout
    stderr = sys.stderr

    sys.argv = [""]
    sys.stdout = io.StringIO()
    sys.stderr = io.StringIO()

    yield

    sys.argv = argv
    sys.stdout = stdout
    sys.stderr = stderr


def makefile(fullpath: str, content: Union[str, bytes] = "", overwrite: bool = False):

    if not os.path.isfile(fullpath):

        dirpath = os.path.dirname(fullpath)

        if not os.path.isdir(dirpath):
            os.makedirs(dirpath)

        _writefile(fullpath, content)

    elif overwrite:

        _writefile(fullpath, content)


def _writefile(fullpath: str, content: Union[str, bytes] = ""):

    if not isinstance(content, (str, bytes)):
        raise TypeError(f"content must be str or bytes, got: {type(content)}")

    with open(fullpath, "wb" if isinstance(content, bytes) else "w") as f:
        f.write(content)



# NOTE: This file is auto-generated by "inv sync_meta"

name = "gutt"
version = "0.6.0"
description = "Auto generate unit test templates"
authors = ["Ryan <ryanchao2012@gmail.com>"]
license = "MIT"
readme = "README.md"
repository = "https://github.com/ryanchao2012/gutt"



import re
from typing import List

import attr

from .parser import split_source_into_blocks

immutable = attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)


def unindent(code: str, level: int = 0, spacing: int = 4) -> str:
    return re.sub(rf"^ {{0,{level * spacing}}}", "", code, flags=re.MULTILINE)


def indent(code: str, level: int = 0, spacing: int = 4) -> str:
    return re.sub(r"^", " " * spacing * level, code, flags=re.MULTILINE)


@immutable
class CodeBlock:
    raw: str
    kind: str = "other"  # NOTE: def, class or other
    name: str = None

    @classmethod
    def collect_from_source(cls, code: str) -> List["CodeBlock"]:

        pat = re.compile(r"^(?P<kind>def|class) (?P<name>\w+)")

        blocks = []
        for bk in split_source_into_blocks(code):

            m = re.match(pat, bk)
            data = m.groupdict() if m else {}

            blocks.append(cls(raw=bk, **data))

        return blocks

    @property
    def children(self, spacing: int = 4) -> List["CodeBlock"]:
        if self.kind != "class":
            return []

        else:
            newline_idx = self.raw.find("\n")

            return self.collect_from_source(
                unindent(
                    self.raw[newline_idx + 1 :], level=1, spacing=spacing  # NOQA: E203
                )
            )



import importlib
import re
from types import ModuleType
from typing import Generator


def load_module_from_pyfile(modname: str, fullpath: str) -> ModuleType:

    spec = importlib.util.spec_from_file_location(modname, fullpath)

    return spec.loader.load_module(modname)


def split_source_into_blocks(source: str) -> Generator[str, None, None]:
    head_index = 0

    pat = re.compile(r"(?=^)[^\s#\'\"\)]+", re.MULTILINE)

    for m in re.finditer(pat, source):
        if m.start() > head_index:
            yield source[head_index : m.start()]  # NOQA: E203

        head_index = m.start()

    yield source[head_index:]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gutt', 'gutt.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'isort<5.9', 'jedi<0.18']

entry_points = \
{'console_scripts': ['gutt = gutt.cli.main:main']}

setup_kwargs = {
    'name': 'gutt',
    'version': '0.6.0',
    'description': 'Auto generate unit test templates',
    'long_description': "# gutt\nAuto Generate Unit Test Templates\n\n## Install\n\n```\n$ pip install gutt\n```\n\n\n## Basic Usage\n\nAssume you have a package, and its layout:\n\n```\nmy_awesome_package\n├── __init__.py\n└── module1.py\n```\n\nsome codes inside `my_awesome_package/module1.py`:\n\n```python\n\nimport sys\n\nMY_CONST = 123\n\ndef funcion1():\n    pass\n\n\ndef function2():\n    pass\n\n\nclass MyObject:\n    def method1(self):\n        pass\n\n    @classmethod\n    def classmethod1(cls):\n        pass\n\n    @staticmethod\n    def staticmethod1():\n        pass\n\n```\n\n`gutt` can generate unit testing templates for all implementations in just one line:\n\n```\n$ gutt -m my_awesome_package -o mytests\n```\n\nThe output layout:\n\n```\nmytests\n├── __init__.py\n└── my_awesome_package\n    ├── __init__.py\n    └── test_module1.py\n\n```\n\nThe unit test templates inside `test_module1.py`\n\n```python\ndef test_funcion1():\n    from my_awesome_package.module1 import funcion1\n\n    assert funcion1\n\n\ndef test_function2():\n    from my_awesome_package.module1 import function2\n\n    assert function2\n\n\nclass TestMyObject:\n    @classmethod\n    def setup_class(cls):\n        from my_awesome_package.module1 import MyObject\n\n        assert MyObject\n\n    @classmethod\n    def teardown_class(cls):\n        pass\n\n    def setup_method(self, method):\n        pass\n\n    def teardown_method(self, method):\n        pass\n\n    def test_method1(self):\n        pass\n\n    def test_classmethod1(self):\n        pass\n\n    def test_staticmethod1(self):\n        pass\n\n```\n\nEach module in source codes maps to a testing module(`module1.py --> test_module1.py`), and each function, each class and all methods inside that class maps to corresponding test templates. \n\n- `gutt` will skip code generation if the test templates for the functions already exist.\n- `gutt` won't delete the corresponding test templates if the source codes get deleted or renamed.\n- For new added codes: modules, functions or methods inside class, just re-run `gutt` to generate new test templates for them.\n\n\nRun unit test with `pytest`, for example:\n\n```\n$ pytest --doctest-modules --cov=my_awesome_package mytests\n\n=============================== test session starts ===============================\nplatform linux -- Python 3.8.8, pytest-4.6.11, py-1.10.0, pluggy-0.13.1\nrootdir: /home/ryan/Workspace/my_awesome_package\nplugins: mock-1.13.0, cov-2.11.1\ncollected 5 items                                                                 \n\nmytests/my_awesome_package/test_module1.py .....                            [100%]\n\n----------- coverage: platform linux, python 3.8.8-final-0 -----------\nName                             Stmts   Miss  Cover\n----------------------------------------------------\nmy_awesome_package/__init__.py       0      0   100%\nmy_awesome_package/module1.py       13      5    62%\n----------------------------------------------------\nTOTAL                               13      5    62%\n\n\n============================ 5 passed in 0.07 seconds =============================\n```",
    'author': 'Ryan',
    'author_email': 'ryanchao2012@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ryanchao2012/gutt',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
